@@ -12,10 +12,11 @@ behave in the same way as the original (at least as closely as possible).
 
 The primary interface to this is the :func:`patch_all` function, which
 performs all the available patches. It accepts arguments to limit the
-patching to certain modules, but most programs will want to use the
-default values as they receive the most wide-spread testing.
+patching to certain modules, but most programs **should** use the
+default values as they receive the most wide-spread testing, and some monkey
+patches have dependencies on others.
 
-Patching *should be done as early as possible* in the lifecycle of the
+Patching **should be done as early as possible** in the lifecycle of the
 program. For example, the main module (the one that tests against
 ``__main__`` or is otherwise the first imported) should begin with
 this code, ideally before any other imports::
@@ -80,14 +81,11 @@ if sys.version_info[0] >= 3:
     string_types = str,
     PY3 = True
 else:
-    import __builtin__
-    string_types = __builtin__.basestring
+    import __builtin__ # pylint:disable=import-error
+    string_types = __builtin__.basestring,
     PY3 = False
 
-if sys.platform.startswith("win"):
-    WIN = True
-else:
-    WIN = False
+WIN = sys.platform.startswith("win")
 
 # maps module name -> {attribute name: original item}
 # e.g. "time" -> {"sleep": built-in function sleep}
@@ -133,19 +131,19 @@ def get_original(mod_name, item_name):
     else:
         return _get_original(mod_name, item_name)
 
+_NONE = object()
+
 
 def patch_item(module, attr, newitem):
-    NONE = object()
-    olditem = getattr(module, attr, NONE)
-    if olditem is not NONE:
+    olditem = getattr(module, attr, _NONE)
+    if olditem is not _NONE:
         saved.setdefault(module.__name__, {}).setdefault(attr, olditem)
     setattr(module, attr, newitem)
 
 
 def remove_item(module, attr):
-    NONE = object()
-    olditem = getattr(module, attr, NONE)
-    if olditem is NONE:
+    olditem = getattr(module, attr, _NONE)
+    if olditem is _NONE:
         return
     saved.setdefault(module.__name__, {}).setdefault(attr, olditem)
     delattr(module, attr)
@@ -220,8 +218,12 @@ def patch_os():
     environment variable ``GEVENT_NOWAITPID`` is not defined). Does
     nothing if fork is not available.
 
-    This method must be used with :func:`patch_signal` to have proper SIGCHLD
-    handling. :func:`patch_all` calls both by default.
+    .. caution:: This method must be used with :func:`patch_signal` to have proper SIGCHLD
+         handling and thus correct results from ``waitpid``.
+         :func:`patch_all` calls both by default.
+
+    .. caution:: For SIGCHLD handling to work correctly, the event loop must run.
+         The easiest way to help ensure this is to use :func:`patch_all`.
     """
     patch_module('os')
 
@@ -247,7 +249,7 @@ def _patch_existing_locks(threading):
         class _ModuleLock(object):
             pass
     else:
-        _ModuleLock = importlib._bootstrap._ModuleLock
+        _ModuleLock = importlib._bootstrap._ModuleLock # python 2 pylint: disable=no-member
     # It might be possible to walk up all the existing stack frames to find
     # locked objects...at least if they use `with`. To be sure, we look at every object
     # Since we're supposed to be done very early in the process, there shouldn't be
@@ -294,6 +296,9 @@ def patch_thread(threading=True, _threading_local=True, Event=False, logging=Tru
     .. versionchanged:: 1.1b1
         Add *logging* and *existing_locks* params.
     """
+    # XXX: Simplify
+    # pylint:disable=too-many-branches,too-many-locals
+
     # Description of the hang:
     # There is an incompatibility with patching 'thread' and the 'multiprocessing' module:
     # The problem is that multiprocessing.queues.Queue uses a half-duplex multiprocessing.Pipe,
@@ -322,22 +327,30 @@ def patch_thread(threading=True, _threading_local=True, Event=False, logging=Tru
     # are in trouble. The latter is tricky because of the different names
     # on different versions.
     if threading:
-        __import__('threading')
+        threading_mod = __import__('threading')
+        # Capture the *real* current thread object before
+        # we start returning DummyThread objects, for comparison
+        # to the main thread.
+        orig_current_thread = threading_mod.current_thread()
+    else:
+        threading_mod = None
+        orig_current_thread = None
 
     patch_module('thread')
+
     if threading:
-        threading = patch_module('threading')
+        patch_module('threading')
 
         if Event:
             from gevent.event import Event
-            patch_item(threading, 'Event', Event)
+            patch_item(threading_mod, 'Event', Event)
 
         if existing_locks:
-            _patch_existing_locks(threading)
+            _patch_existing_locks(threading_mod)
 
         if logging and 'logging' in sys.modules:
             logging = __import__('logging')
-            patch_item(logging, '_lock', threading.RLock())
+            patch_item(logging, '_lock', threading_mod.RLock())
             for wr in logging._handlerList:
                 # In py26, these are actual handlers, not weakrefs
                 handler = wr() if callable(wr) else wr
@@ -345,13 +358,43 @@ def patch_thread(threading=True, _threading_local=True, Event=False, logging=Tru
                     continue
                 if not hasattr(handler, 'lock'):
                     raise TypeError("Unknown/unsupported handler %r" % handler)
-                handler.lock = threading.RLock()
+                handler.lock = threading_mod.RLock()
 
     if _threading_local:
         _threading_local = __import__('_threading_local')
         from gevent.local import local
         patch_item(_threading_local, 'local', local)
 
+    def make_join_func(thread, thread_greenlet):
+        from gevent.hub import sleep
+        from time import time
+
+        def join(timeout=None):
+            end = None
+            if threading_mod.current_thread() is thread:
+                raise RuntimeError("Cannot join current thread")
+            if thread_greenlet is not None and thread_greenlet.dead:
+                return
+            if not thread.is_alive():
+                return
+
+            if timeout:
+                end = time() + timeout
+
+            while thread.is_alive():
+                if end is not None and time() > end:
+                    return
+                sleep(0.01)
+        return join
+
+    if threading:
+        from gevent.threading import main_native_thread
+
+        for thread in threading_mod._active.values():
+            if thread == main_native_thread():
+                continue
+            thread.join = make_join_func(thread, None)
+
     if sys.version_info[:2] >= (3, 4):
 
         # Issue 18808 changes the nature of Thread.join() to use
@@ -359,36 +402,25 @@ def patch_thread(threading=True, _threading_local=True, Event=False, logging=Tru
         # (which is already running) cannot wait for the main thread---it
         # hangs forever. We patch around this if possible. See also
         # gevent.threading.
-        threading = __import__('threading')
         greenlet = __import__('greenlet')
-        if threading.current_thread() == threading.main_thread():
-            main_thread = threading.main_thread()
+
+        if orig_current_thread == threading_mod.main_thread():
+            main_thread = threading_mod.main_thread()
             _greenlet = main_thread._greenlet = greenlet.getcurrent()
             from gevent.hub import sleep
 
-            def join(timeout=None):
-                if threading.current_thread() is main_thread:
-                    raise RuntimeError("Cannot join current thread")
-                if _greenlet.dead or not main_thread.is_alive():
-                    return
-                elif timeout:
-                    raise ValueError("Cannot use a timeout to join the main thread")
-                    # XXX: Make that work
-                else:
-                    while main_thread.is_alive():
-                        sleep(0.01)
 
-            main_thread.join = join
+            main_thread.join = make_join_func(main_thread, _greenlet)
 
             # Patch up the ident of the main thread to match. This
             # matters if threading was imported before monkey-patching
             # thread
             oldid = main_thread.ident
-            main_thread._ident = threading.get_ident()
-            if oldid in threading._active:
-                threading._active[main_thread.ident] = threading._active[oldid]
+            main_thread._ident = threading_mod.get_ident()
+            if oldid in threading_mod._active:
+                threading_mod._active[main_thread.ident] = threading_mod._active[oldid]
             if oldid != main_thread.ident:
-                del threading._active[oldid]
+                del threading_mod._active[oldid]
         else:
             _queue_warning("Monkey-patching not on the main thread; "
                            "threading.main_thread().join() will hang from a greenlet",
@@ -407,12 +439,12 @@ def patch_socket(dns=True, aggressive=True):
     # However, because gevent.socket.socket.connect is a Python function, the exception raised by it causes
     # _socket object to be referenced by the frame, thus causing the next invocation of bind(source_address) to fail.
     if dns:
-        items = socket.__implements__
+        items = socket.__implements__ # pylint:disable=no-member
     else:
-        items = set(socket.__implements__) - set(socket.__dns__)
+        items = set(socket.__implements__) - set(socket.__dns__) # pylint:disable=no-member
     patch_module('socket', items=items)
     if aggressive:
-        if 'ssl' not in socket.__implements__:
+        if 'ssl' not in socket.__implements__: # pylint:disable=no-member
             remove_item(socket, 'ssl')
 
 
@@ -423,7 +455,7 @@ def patch_dns():
     by that method if requested.
     """
     from gevent import socket
-    patch_module('socket', items=socket.__dns__)
+    patch_module('socket', items=socket.__dns__) # pylint:disable=no-member
 
 
 def patch_ssl():
@@ -524,8 +556,11 @@ def patch_signal():
     """
     Make the signal.signal function work with a monkey-patched os.
 
-    This method must be used with :func:`patch_os` to have proper SIGCHLD
-    handling. :func:`patch_all` calls both by default.
+    .. caution:: This method must be used with :func:`patch_os` to have proper SIGCHLD
+         handling. :func:`patch_all` calls both by default.
+
+    .. caution:: For proper SIGCHLD handling, you must yield to the event loop.
+         Using :func:`patch_all` is the easiest way to ensure this.
 
     .. seealso:: :mod:`gevent.signal`
     """
@@ -561,6 +596,8 @@ def patch_all(socket=True, dns=True, time=True, select=True, thread=True, os=Tru
        and ``signal=True``. This will cause SIGCHLD handlers to not be called. This may
        be an error in the future.
     """
+    # pylint:disable=too-many-locals,too-many-branches
+
     # Check to see if they're changing the patched list
     _warnings, first_time = _check_repatching(**locals())
     if not _warnings and not first_time:
@@ -574,7 +611,7 @@ def patch_all(socket=True, dns=True, time=True, select=True, thread=True, os=Tru
     if time:
         patch_time()
     if thread:
-        patch_thread(Event=Event)
+        patch_thread(Event=Event, _warnings=_warnings)
     # sys must be patched after thread. in other cases threading._shutdown will be
     # initiated to _MainThread with real thread ident
     if sys:
@@ -646,7 +683,7 @@ def main():
 
 def _get_script_help():
     from inspect import getargspec
-    patch_all_args = getargspec(patch_all)[0]
+    patch_all_args = getargspec(patch_all)[0] # pylint:disable=deprecated-method
     modules = [x for x in patch_all_args if 'patch_' + x in globals()]
     script_help = """gevent.monkey - monkey patch the standard modules to use gevent.
 
